from abc import ABCMeta, abstractmethod
from torchqtm.utils.universe import Universe
import pandas as pd
import numpy as np
from multiprocessing import Pool
from typing import Dict, Hashable
from torchqtm.base import BackTestEnv
import matplotlib.pyplot as plt
from torchqtm.utils.visualization import ColorGenerator
import torchqtm.op.functional as F
from typing import final


class BaseTester(object, metaclass=ABCMeta):
    def __init__(self,
                 env: BackTestEnv = None):
        self.metrics = None
        self.results = None
        self.env = env
        self._check_env()
        self.rebalance_dates = env.dates

    def _check_env(self):
        ...
        # assert '_FutureReturn' in self.env
        # assert 'MktVal' in self.env
        # assert 'Sector' in self.env
        # assert 'Close' in self.env

    def _reset(self):
        self.metrics = None
        self.results = None

    @abstractmethod
    def run_backtest(self, modified_factor) -> None:
        raise NotImplementedError("Should implement in the derived class.")


class TesterMixin:
    def score(self, alpha, method='pearson'):
        return np.nanmean(F.cs_corr(alpha, self.env.create_forward_returns(), method=method), axis=0)

    # def _more_tags(self):
    #     return {"requires_y": True}


class BaseGroupTester(BaseTester, TesterMixin):
    def __init__(self,
                 env: BackTestEnv = None,
                 n_groups: int = 5,
                 weighting: str = 'equal',
                 exclude_suspended: bool = False,
                 exclude_limits: bool = False):
        """

        Parameters
        ----------
        env
        n_groups
        weighting: "equal" or "market_cap" or "factor_cap"
        exclude_suspended
        exclude_limits
        """
        super().__init__(env)
        self.n_groups = n_groups
        self.weighting = weighting
        self.exclude_limits = exclude_limits
        self.exclude_suspended = exclude_suspended
        self.returns = None

    def _reset(self):
        super()._reset()
        self.returns = None

    @abstractmethod
    def run_backtest(self, modified_factor) -> None:
        raise NotImplementedError("Should implement in the derived class.")

    # TODO: add more parameters
    def plot(self):
        # plot the result
        fig = plt.figure(figsize=(20, 12))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        color_generator = ColorGenerator(self.n_groups)
        colors = color_generator.data
        for i in range(self.n_groups):
            ax.plot((1 + self.returns.iloc[:, i]).cumprod(), label=f'group_{i + 1}', color=colors[i])
        fig.legend(fontsize=16)
        fig.show()


class BaseIcTester(BaseTester, TesterMixin):
    def __init__(self,
                 env: BackTestEnv = None,
                 universe: Universe = None,
                 method: str = "pearson"):
        super().__init__(env, universe)
        self.method = method
        self.results = None

    def _reset(self):
        super()._reset()

    @abstractmethod
    def run_backtest(self, modified_factor) -> None:
        raise NotImplementedError
        # self._reset()
        # self.results = F.cs_corr(modified_factor, self.env._FutureReturn)
        # self.metrics = np.nanmean(self.results, axis=0)

    @final
    def plot(self):
        fig = plt.figure(figsize=(20, 12))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.plot(self.rebalance_dates, self.results)
        fig.legend(fontsize=16)
        fig.show()


class GroupTester01(BaseGroupTester):
    def __init__(self,
                 env: BackTestEnv = None,
                 n_groups: int = 5,
                 weighting: str = 'equal',
                 exclude_suspended: bool = False,
                 exclude_limits: bool = False):
        super().__init__(env, n_groups, weighting, exclude_suspended, exclude_limits)

    def run_backtest(self,
                     modified_factor,
                     purify=True) -> None:
        assert modified_factor.shape == self.env['Close'].shape
        if purify:
            modified_factor = F.purify(modified_factor)
        self._reset()
        labels = ["group_" + str(i + 1) for i in range(self.n_groups)]
        returns = []
        forward_return = self.env.create_forward_returns(D=1)
        symbols = list(self.env.Close.columns)

        datas = np.concatenate([forward_return.values[..., np.newaxis],
                                self.env.MktVal.values[..., np.newaxis],
                                modified_factor.values[..., np.newaxis]], axis=2)

        for i in range(len(modified_factor)):

            data_slice = pd.DataFrame(datas[i],
                                      index=symbols,
                                      columns=['forward_returns', 'MktVal', 'modified_factor'])
            data_masked = data_slice.loc[~np.isnan(data_slice['modified_factor'])].copy()
            if len(data_masked) == 0 or data_masked['forward_returns'].isna().all():
                group_return = pd.Series(0, index=labels)
            else:
                data_masked.loc[:, 'group'] = pd.qcut(data_masked['modified_factor'], self.n_groups, labels=labels)

                def temp(x):
                    # TODO: develop a weight_scheme class
                    if self.weighting == 'equal':
                        weight = 1 / len(x['MktVal'])
                    elif self.weighting == 'market_cap':
                        weight = x['MktVal'] / x['MktVal'].sum()
                    elif self.weighting == 'factor_cap':
                        weight = x['modified_factor'] / np.sum(np.abs(x['modified_factor']))
                    else:
                        raise ValueError('Invalid weight scheme')
                    ret = x['forward_returns']
                    return (weight * ret).sum()

                group_return = data_masked.groupby('group').apply(temp)
            returns.append(group_return)
        self.returns = pd.concat(returns, axis=1).T
        # Here we need to transpose the return, since the rows are stocks.
        self.returns.index = self.rebalance_dates
        self.returns.index.name = "trade_date"
        self.returns.columns.name = "group"


class LongShort01(BaseGroupTester):
    def __init__(self,
                 env: BackTestEnv = None,
                 n_groups: int = 5,
                 leverage: float = 1,
                 weighting: str = 'equal',
                 exclude_suspended: bool = False,
                 exclude_limits: bool = False):
        super().__init__(env, n_groups, weighting, exclude_suspended, exclude_limits)
        self.leverage = leverage

    def run_backtest(self,
                     modified_factor,
                     purify=True) -> None:
        """
        self.returns
        """
        assert modified_factor.shape == self.env['Close'].shape
        if purify:
            modified_factor = F.purify(modified_factor)
        self._reset()
        labels = ["group_" + str(i + 1) for i in range(self.n_groups)]
        returns = []
        forward_return = self.env.create_forward_returns(D=1)
        symbols = list(self.env.Close.columns)

        datas = np.concatenate([forward_return.values[..., np.newaxis],
                                self.env.MktVal.values[..., np.newaxis],
                                modified_factor.values[..., np.newaxis]], axis=2)

        for i in range(len(modified_factor)):

            data_slice = pd.DataFrame(datas[i],
                                      index=symbols,
                                      columns=['forward_returns', 'MktVal', 'modified_factor'])
            data_masked = data_slice.loc[~np.isnan(data_slice['modified_factor'])].copy()
            if len(data_masked) == 0 or data_masked['forward_returns'].isna().all():
                group_return = pd.Series(0, index=labels)
            else:
                data_masked.loc[:, 'group'] = pd.qcut(data_masked['modified_factor'], self.n_groups, labels=labels)

                def aggHelper(x):
                    # TODO: develop a weight_scheme class
                    if self.weighting == 'equal':
                        weight = 1 / len(x['MktVal'])
                    elif self.weighting == 'market_cap':
                        weight = x['MktVal'] / x['MktVal'].sum()
                    elif self.weighting == 'factor_cap':
                        weight = x['modified_factor'] / np.sum(np.abs(x['modified_factor']))
                    else:
                        raise ValueError('Invalid weight scheme')
                    ret = x['forward_returns']
                    return (weight * ret).sum()

                group_return = data_masked.groupby('group').apply(aggHelper)
            returns.append(group_return)
        temp_data = pd.concat(returns, axis=1).T
        returns = temp_data[f'group_{self.n_groups+1}'] - temp_data[f'group_{1}']
        if self.leverage == 1:
            self.returns = returns / 2
        else:
            self.returns = returns
        # Here we need to transpose the return, since the rows are stocks.
        self.returns.index = self.rebalance_dates

    @final
    def plot(self):
        fig = plt.figure(figsize=(20, 12))
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.plot((1 + self.returns).cumprod())
        fig.legend(fontsize=16)
        fig.show()


class WQTester(BaseTester, TesterMixin):

    def __init__(self,
                 env: BackTestEnv = None,
                 exclude_suspended: bool = False,
                 exclude_limits: bool = False
                 ):
        super().__init__(env)
        self.env = env
        self.exclude_suspended = exclude_suspended
        self.returns = None
        self.exclude_limits = exclude_limits

    def _reset(self):
        super()._reset()
        self.returns = None

    def run_backtest(self,
                     modified_factor,
                     purify=True):

        assert modified_factor.shape == self.env['Close'].shape
        if purify:
            modified_factor = F.purify(modified_factor)
        self._reset()
        forward_return = self.env.create_forward_returns(D=1)
        symbols = list(self.env.Close.columns)
        returns = []
        datas = np.concatenate([forward_return.values[..., np.newaxis],
                                self.env.MktVal.values[..., np.newaxis],
                                modified_factor.values[..., np.newaxis]], axis=2)

        for i in range(len(modified_factor)):

            # Step1: Evaluate the expression for each stock to generate the alpha vector for the given date.
            data_slice = pd.DataFrame(datas[i],
                                      index=symbols,
                                      columns=['forward_returns', 'MktVal', 'modified_factor'])
            data_masked = data_slice.loc[~np.isnan(data_slice['modified_factor'])].copy()

            # Step2: From each value in the vector, subtract the average of the vector values in the group.
            # Sum of all vector values = 0. This is called neutralization.
            # The group can be the entire market,
            # but we can also perform this neutralization operation on sector,
            # industry or subindustry groupings of stocks.
            data_masked['modified_factor'] -= np.mean(data_masked['modified_factor'])

            # Step 3: The resulting values are scaled or ‘normalized’ such that absolute sum of the alpha vector
            # values is 1. These values can be called as normalized weights.
            data_masked['modified_factor'] /= np.sum(np.abc(data_masked['modified_factor']))

            # Step 4: Using normalized weights, the BRAIN simulator allocates capital
            # (from a fictitious book of $20 million) to each stock to construct a portfolio.
            returns.append(np.sum(data_masked['modified_factor'] * data_masked['forward_returns']))

        self.returns = pd.Series(returns, index=self.rebalance_dates)



import pandas as pd

start = pd.Timestamp('1990-01-01')
end_base = pd.Timestamp('today')
# Give an aggressive buffer for logic that needs to use the next trading
# day or minute.
end = end_base + pd.Timedelta(days=365)





"""The functions used to create programs.

The :mod:`gplearn.functions` module contains all of the functions used by
gplearn programs. It also contains helper methods for a user to define their
own custom functions.
"""

# Author: Trevor Stephens <trevorstephens.com>
#
# License: BSD 3 clause

import numpy as np
import pandas as pd
from joblib import wrap_non_picklable_objects
import torchqtm.op.functional as F
import functools

__all__ = ['make_function']


class _Function(object):
    """A representation of a mathematical relationship, a node in a program.

    This object is able to be called with NumPy torchqtm arguments and return
    a resulting vector based on a mathematical relationship.

    Parameters
    ----------
    function : callable
        A function with signature function(x1, *args) that returns a Numpy
        array of the same shape as its arguments.

    name : str
        The name for the function as it should be represented in the program
        and its visualizations.

    arity : int
        The number of arguments that the ``function`` takes.

    """

    def __init__(self, function, name, arity):
        self.function = function
        self.name = name
        self.arity = arity

    def __call__(self, *args):
        return self.function(*args)


def make_function(*, function, name, arity, wrap=True):
    """Make a function node, a representation of a mathematical relationship.

    This factory function creates a function node, one of the core nodes in any
    program. The resulting object is able to be called with NumPy torchqtm
    arguments and return a resulting vector based on a mathematical
    relationship.

    Parameters
    ----------
    function : callable
        A function with signature `function(x1, *args)` that returns a Numpy
        array of the same shape as its arguments.

    name : str
        The name for the function as it should be represented in the program
        and its visualizations.

    arity : int
        The number of arguments that the `function` takes.

    wrap : bool, optional (default=True)
        When running in parallel, pickling of custom functions is not supported
        by Python's default pickler. This option will wrap the function using
        cloudpickle allowing you to pickle your solution, but the evolution may
        run_ slightly more slowly. If you are running single-threaded in an
        interactive Python session or have no need to save the model, set to
        `False` for faster runs.

    """
    if not isinstance(arity, int):
        raise ValueError('arity must be an int, got %s' % type(arity))
    if not isinstance(function, np.ufunc):
        if function.__code__.co_argcount != arity:
            raise ValueError('arity %d does not match required number of '
                             'function arguments of %d.'
                             % (arity, function.__code__.co_argcount))
    if not isinstance(name, str):
        raise ValueError('name must be a string, got %s' % type(name))
    if not isinstance(wrap, bool):
        raise ValueError('wrap must be an bool, got %s' % type(wrap))

    # Check output shape
    args = [np.ones(10) for _ in range(arity)]
    try:
        function(*args)
    except (ValueError, TypeError):
        raise ValueError('supplied function %s does not support arity of %d.'
                         % (name, arity))
    if not hasattr(function(*args), 'shape'):
        raise ValueError('supplied function %s does not return a numpy array.'
                         % name)
    if function(*args).shape != (10,):
        raise ValueError('supplied function %s does not return same shape as '
                         'input vectors.' % name)

    # Check closure for zero & negative input arguments
    args = [np.zeros(10) for _ in range(arity)]
    if not np.all(np.isfinite(function(*args))):
        raise ValueError('supplied function %s does not have closure against '
                         'zeros in argument vectors.' % name)
    args = [-1 * np.ones(10) for _ in range(arity)]
    if not np.all(np.isfinite(function(*args))):
        raise ValueError('supplied function %s does not have closure against '
                         'negatives in argument vectors.' % name)

    if wrap:
        return _Function(function=wrap_non_picklable_objects(function),
                         name=name,
                         arity=arity)
    return _Function(function=function,
                     name=name,
                     arity=arity)


def _protected_division(x1, x2):
    """Closure of division (x1/x2) for zero denominator."""
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.where(np.abs(x2) > 0.001, np.divide(x1, x2), 1.)


def _protected_sqrt(x1):
    """Closure of square root for negative arguments."""
    return np.sqrt(np.abs(x1))


def _protected_log(x1):
    """Closure of log for zero and negative arguments."""
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.where(np.abs(x1) > 0.001, np.log(np.abs(x1)), 0.)


def _protected_inverse(x1):
    """Closure of inverse for zero arguments."""
    with np.errstate(divide='ignore', invalid='ignore'):
        return np.where(np.abs(x1) > 0.001, 1. / x1, 0.)


def _sigmoid(x1):
    """Special case of logistic function to transform to probabilities."""
    with np.errstate(over='ignore', under='ignore'):
        return 1 / (1 + np.exp(-x1))


add2 = _Function(function=np.add, name='add', arity=2)
sub2 = _Function(function=np.subtract, name='sub', arity=2)
mul2 = _Function(function=np.multiply, name='mul', arity=2)
div2 = _Function(function=_protected_division, name='div', arity=2)
sqrt1 = _Function(function=_protected_sqrt, name='sqrt', arity=1)
log1 = _Function(function=_protected_log, name='log', arity=1)
neg1 = _Function(function=np.negative, name='neg', arity=1)
inv1 = _Function(function=_protected_inverse, name='inv', arity=1)
abs1 = _Function(function=np.abs, name='abs', arity=1)
max2 = _Function(function=np.maximum, name='max', arity=2)
min2 = _Function(function=np.minimum, name='min', arity=2)
sin1 = _Function(function=np.sin, name='sin', arity=1)
cos1 = _Function(function=np.cos, name='cos', arity=1)
tan1 = _Function(function=np.tan, name='tan', arity=1)
sig1 = _Function(function=_sigmoid, name='sig', arity=1)


def correlation(x, y, d):
    """
    :param x:
    :param y:
    :param d:
    :return:
    """
    x = pd.Series(x)
    y = pd.Series(y)
    return x.rolling(d).corr(y).fillna(0)


correlation5 = functools.partial(correlation, d=5)
correlation5 = _Function(function=correlation5, name='correlation5', arity=2)
correlation10 = functools.partial(correlation, d=10)
correlation10 = _Function(function=correlation10, name='correlation10', arity=2)


_function_map = {'add': add2,
                 'sub': sub2,
                 'mul': mul2,
                 'div': div2,
                 'sqrt': sqrt1,
                 'log': log1,
                 'abs': abs1,
                 'neg': neg1,
                 'inv': inv1,
                 'max': max2,
                 'min': min2,
                 'sin': sin1,
                 'cos': cos1,
                 'tan': tan1}



"""Genetic Programming in Python, with a scikit-learn inspired API

``gplearn`` is a set of algorithms for learning genetic programming models.

"""
__version__ = '0.4.2'

__all__ = ['genetic', 'functions', 'fitness']



import numpy as np
from scipy.stats import norm
from scipy.optimize import brentq
from functools import partial
import math
from scipy.stats import norm


def bsm_call(T, S, K, r, sigma):
    d1 = (np.log(S / K) + (r + 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))
    d2 = (np.log(S / K) + (r - 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))

    call = (S * norm.cdf(d1, 0.0, 1.0) - K * np.exp(-r * T) * norm.cdf(d2, 0.0, 1.0))
    return call


def bsm_call_delta(T, S, K, r, sigma):
    """
    delta = \frac{\ln{\frac{S}{K}} + (r + \frac{1}{2} \sigma^2 }{\sigma * \sqrt{T}}
    """
    d1 = (np.log(S / K) + (r + 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))
    return norm.cdf(d1, 0.0, 1)


def bsm_call_gamma(T, S, K, r, sigma):
    pass


def bsm_call_vega(T, S, K, r, sigma):
    """
    vega = \f
    judgement: the at the money strike has the lowest vega(X).
    """
    d1 = (np.log(S / K) + (r + 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))
    return S * norm.pdf(d1, 0.0, 1) * np.sqrt(T)


def _Newton(Df, DDf, x_0=0, N=100):
    rlt = [x_0] * N
    if DDf is None:
        for i in range(N):
            rlt[i] = rlt[i - 1] - Df(rlt[i - 1]) * (-0.01)
    else:
        for i in range(N):
            rlt[i] = rlt[i - 1] - Df(rlt[i - 1]) / DDf(rlt[i - 1])
    return rlt[-1]


def implied_vol(T, S, K, r, price, method='brentq'):
    if method == 'brentq':
        rlt = brentq(lambda x: price - bsm_call(T, S, K, r, x), 1e-6, 1)
    elif method == 'newton':
        df = lambda x: price - bsm_call(T, S, K, r, x)
        ddf = lambda x: -bsm_call_vega(T, S, K, r, x)
        rlt = _Newton(df, None, x_0=1, N=1000)
    else:
        raise ValueError
    return rlt


def sabr(alpha, beta, rho, nu, F, K, T, option='C'):
    X = K
    # if K is at-the-money-forward
    if abs(F - K) < 1e-12:
        numer1 = (((1 - beta)**2)/24)*alpha*alpha/(F**(2 - 2*beta))
        numer2 = 0.25*rho*beta*nu*alpha/(F**(1 - beta))
        numer3 = ((2 - 3*rho*rho)/24)*nu*nu
        VolAtm = alpha*(1 + (numer1 + numer2 + numer3)*T)/(F**(1-beta))
        sabrsigma = VolAtm
    else:
        z = nu/alpha*math.pow(F*X, (1-beta)/2)*math.log(F/X)
        zhi = math.log((math.sqrt(1 - 2*rho*z + z*z) + z - rho)/(1 - rho))
        numer1 = (((1 - beta)**2)/24)*((alpha*alpha)/((F*X)**(1 - beta)))
        numer2 = 0.25*rho*beta*nu*alpha/((F*X)**((1 - beta)/2))
        numer3 = ((2 - 3*rho*rho)/24)*nu*nu
        numer = alpha*(1 + (numer1 + numer2 + numer3)*T)*z
        denom1 = ((1 - beta)**2/24)*(math.log(F/X))**2
        denom2 = (((1 - beta)**4)/1920)*((math.log(F/X))**4)
        denom = ((F*X)**((1 - beta)/2))*(1 + denom1 + denom2)*zhi
        sabrsigma = numer/denom

    return sabrsigma



import pandas as pd
import numpy as np

import datetime

from torchqtm.constants import (
    GLOBAL_DEFAULT_MARKET_OPEN,
    GLOBAL_DEFAULT_MARKET_CLOSE,
    GLOBAL_DEFAULT_BEFORE_TRADING_START_OFFSET,
)

from torchqtm.types import EVENT_TYPE


class DailySimulationClock(object):
    def __init__(
            self,
            sessions: pd.DatetimeIndex,
            market_open: datetime.time = GLOBAL_DEFAULT_MARKET_OPEN,
            market_close: datetime.time = GLOBAL_DEFAULT_MARKET_CLOSE,
            offset: int = GLOBAL_DEFAULT_BEFORE_TRADING_START_OFFSET,
    ):
        self.sessions_nanos = sessions.values.astype(np.int64)
        duration_0 = (market_open.hour * 60 + market_open.minute - offset) * 60 * 1e9
        duration_1 = (market_close.hour * 60 + market_close.minute) * 60 * 1e9
        self.bts_nanos = self.sessions_nanos + duration_0
        self.market_close_nanos = self.sessions_nanos + duration_1

    def __iter__(self):
        for idx in range(len(self.sessions_nanos)):
            session_nano = self.sessions_nanos[idx]
            bts_nano = self.bts_nanos[idx]
            market_close_nano = self.market_close_nanos[idx]
            yield pd.Timestamp(session_nano), EVENT_TYPE.SESSION_START
            yield pd.Timestamp(bts_nano), EVENT_TYPE.BEFORE_TRADING_START
            yield pd.Timestamp(market_close_nano), EVENT_TYPE.BAR
            yield pd.Timestamp(market_close_nano), EVENT_TYPE.SESSION_END


class MinuteSimulationClock(object):
    def __init__(self,
                 sessions: pd.DatetimeIndex,
                 market_open: datetime.time,
                 market_close: datetime.time,
                 offset: int):
        self.sessions_nanos = sessions.values.astype(np.int64)
        duration_0 = (market_open.hour * 60 + market_open.minute - offset) * 60 * 1e9
        duration_1 = (market_close.hour * 60 + market_close.minute) * 60 * 1e9
        self.bts_nanos = self.sessions_nanos + duration_0
        self.market_close_nanos = self.sessions_nanos + duration_1

    def __iter__(self):
        # for idx in range(len(self.sessions_nanos)):
        #     session_nano = self.sessions_nanos[idx]
        #     bts_nano = self.bts_nanos[idx]
        #     market_close_nano = self.market_close_nanos[idx]
        #     yield pd.Timestamp(session_nano), Event.SESSION_START
        #     yield pd.Timestamp(bts_nano), Event.BEFORE_TRADING_START
        #     yield pd.Timestamp(market_close_nano), Event.BAR
        #     yield pd.Timestamp(market_close_nano), Event.SESSION_END
        pass



from abc import ABCMeta, abstractmethod
import datetime
import logging
import sys
import pandas as pd
import numpy as np

from torchqtm.edbt.sim_params import SimulationParameters
from torchqtm.finance.metrics.tracker import MetricsTracker
from torchqtm.finance.orders.order import Order
from torchqtm.finance.orders.tracker import OrdersTracker
from torchqtm.assets import Asset, Equity, Future
from torchqtm.data.data_portal import DataPortal
from torchqtm.edbt.gens.sim_engine import DailySimulationClock, MinuteSimulationClock
import typing
from torchqtm.types import EVENT_TYPE, ASSET_TYPE, DATA_FREQUENCIES
from torchqtm.finance.account import Account
from torchqtm.utils.exchange_calendar import XSHGExchangeCalendar
from torchqtm.utils.datetime_utils import DateTimeManager, DateTimeMixin


class TradingAlgorithm(DateTimeMixin):
    _allow_setattr: bool = False

    def __init__(
            self,
            sim_params: SimulationParameters,
            data_portal: DataPortal,
            namespace=None,
            trading_calendar: XSHGExchangeCalendar = None,
            datetime_manager: DateTimeManager = None,
            benchmark_returns=None,
            account: Account = None,
            metrics_tracker: MetricsTracker = None,
    ):
        self._allow_setattr = True
        # Create a reference to the data_portal datetime
        DateTimeMixin.__init__(self, datetime_manager)
        assert account is not None
        assert metrics_tracker is not None

        self.data_portal = data_portal

        self.benchmark_returns = benchmark_returns

        self.sim_params = sim_params

        self.trading_calendar = trading_calendar

        self._last_sync_time = None
        self.metrics_tracker = metrics_tracker

        # if self._metrics_set is None:
        #     self._metrics_set = load_metrics_set("default")

        # TODO: initialize this, 我认为不需要创建account, 只需要创建account的引用就行了
        self.account = account

        self.initialized = False

        self.capital_changes = {}
        self.recorded_vars = {}

        # self.restrictions = NoRestrictions()
        self._allow_setattr = False

    # def __setattr__(self, key, value):
    #     if self._allow_setattr or hasattr(self, key):
    #         # 思考一下用super().__setattr__好还是setattr好, 后者可能会造成递归
    #         super().__setattr__(key, value)
    #     else:
    #         raise AttributeError(f"You cannot add attributes to {self.__class__.__name__}")

    def safe_set_attr(self, attr, value):
        """
        Safely sets an attribute to the object.

        This method checks whether an attribute with the same name already exists on the object.
        If it does, it raises an AttributeError. Otherwise, it sets the attribute to the given value.

        Parameters:
            attr (str): The name of the attribute to set.
            value: The value to set the attribute to.

        Raises:
            AttributeError: If an attribute with the same name already exists on the object.
        """
        if hasattr(self, attr):
            raise AttributeError(f"Attribute {attr} already exists and cannot be set")
        else:
            self._allow_setattr = True
            try:
                setattr(self, attr, value)
            finally:
                self._allow_setattr = False

    @abstractmethod
    def initialize(self):
        """
        Set the initial params of backtest.

        Note:
            use self.safe_set_attr('i', 0) instead of self.i = 0 to avoid overwriting the existing attributes
        """
        raise NotImplementedError("should be implemented in the derived class")

    def _initialize(self):
        self.initialize()

    def before_trading_start(self):
        # self.compute_eager_pipelines()
        pass

    def _before_trading_start(self):
        """
        总体上与api保持一致, 但我用来控制一些内部行为了
        """
        self.before_trading_start()

    def handle_data(self):
        pass

    def _handle_data(self):
        # self.rule.should_trigger(self.current_dt) # FIXME
        self.handle_data()

    def analyze(self):
        pass

    def _analyze(self):
        self.analyze()

    # def __repr__(self):
    #     pass

    def _create_sim_engine(self):
        self.on_dt_changed(self.sim_params.start_session)

        if not self.initialized:
            # TODO: do something
            self.initialized = True

        # benchmark_source = self._create_benchmark_source()

        return DailySimulationClock(
            sessions=self.sim_params.sessions,
            market_open=self.trading_calendar.market_open_time,
            market_close=self.trading_calendar.market_close_time,
        )

    def _create_benchmark_source(self):
        return None

    def _create_metrics_tracker(self):
        return None

    def _create_generator(self, sim_params):

        sim_engine = self._create_sim_engine()

        # benchmark_source = self._create_benchmark_source()

        algo_iter = AlgorithmIterator(
            algo=self,
            datetime_manager=self.datetime_manager,
            sim_params=self.sim_params,
            data_portal=self.data_portal,
            sim_engine=sim_engine,
            benchmark_source=None, # TODO: fixme
            restrictions=None,
        )
        return algo_iter
        # metrics_tracker.handle_start_of_simulation(benchmark_source)

    # def compute_eager_pipelines(self):
    #     raise NotImplementedError

    def get_generator(self):
        self._initialize()
        return self._create_generator(self.sim_params)

    def _create_daily_stats(self):
        pass

    def calculate_capital_changes(
            self,
            dt: pd.Timestamp,
            emission_rate: typing.Literal["daily", "minute"],
            is_interday: bool,
            portfolio_value_adjustment: float = 0.0,
    ):
        """
        在session start的时候调用, 计算资本利得
        """
        try:
            capital_change = self.capital_changes[dt]
        except KeyError:
            return

        self._sync_last_sale_prices()


    def run(self):
        # try:
        #     perfs = []
        #     it = self.get_generator()
        # while True:
        #     try:
        #         perf = next(it)
        #         perfs.append(perf)
        #     except StopIteration:
        #         break
        #
        # daily_stats = self._create_daily_stats(perfs)
        # self._analyze(daily_stats)
        # return daily_stats
        pass

    def record(self):
        """
        Track and record values each day
        """
        pass

    def symbol(self):
        pass

    def symbols(self):
        pass

    def _can_order_asset(self, asset) -> bool:
        pass


    #TODO: revise it
    @staticmethod
    def round(amount: float) -> int:
        return int(amount)

    def validate_order_params(self, asset, amount, limit_price, stop_price):
        # for control in self.trading_controls:
        #     for control in self.trading_controls:
        #         control.validate(
        #             asset,
        #             amount,
        #             self.portfolio,
        #             self.get_datetime(),
        #             self.trading_client.current_data,
        #         )
        # return None
        return None

    def _calculate_order_value_amount(self, asset, value):
        # """Calculates how many shares/contracts to order based on the type of
        # asset being ordered.
        # """
        # # Make sure the asset exists, and that there is a last price for it.
        # # FIXME: we should use BarData's can_trade logic here, but I haven't
        # # yet found a good way to do that.
        # normalized_date = self.trading_calendar.minute_to_session(self.datetime)
        #
        # if normalized_date < asset.start_date:
        #     raise CannotOrderDelistedAsset(
        #         msg="Cannot order {0}, as it started trading on"
        #         " {1}.".format(asset.symbol, asset.start_date)
        #     )
        # elif normalized_date > asset.end_date:
        #     raise CannotOrderDelistedAsset(
        #         msg="Cannot order {0}, as it stopped trading on"
        #         " {1}.".format(asset.symbol, asset.end_date)
        #     )
        # else:
        #     last_price = self.trading_client.current_data.current(asset, "price")
        #
        #     if np.isnan(last_price):
        #         raise CannotOrderDelistedAsset(
        #             msg="Cannot order {0} on {1} as there is no last "
        #             "price for the security.".format(asset.symbol, self.datetime)
        #         )
        #
        # if tolerant_equals(last_price, 0):
        #     zero_message = "Price of 0 for {psid}; can't infer value".format(psid=asset)
        #     if self.logger:
        #         self.logger.debug(zero_message)
        #     # Don't place any order
        #     return 0
        #
        # value_multiplier = asset.price_multiplier
        #
        # return value / (last_price * value_multiplier)
        pass

    def order(self, asset, amount, limit_price=None, stop_price=None):
        self.validate_order_params(asset, amount, limit_price, stop_price)
        self.account.order(asset, amount, limit_price, stop_price)

    def order_value(self, asset, value, limit_price=None, stop_price=None):
        if not self._can_order_asset(asset):
            return None
        amount = self._calculate_order_value_amount(asset, value)
        return self.order(asset, amount, limit_price, stop_price)

    def order_percent(self, asset, percent, limit_price=None, stop_price=None):
        pass

    def order_target(self, asset, target, limit_price=None, stop_price=None):
        pass

    def order_target_value(self, asset, target_value, limit_price=None, stop_price=None):
        pass

    def order_target_percent(self, asset, target_percent, limit_price=None, stop_price=None):
        pass

    def _sync_last_sale_prices(self, dt=None):
        if dt is None:
            dt = self.current_dt

        if dt != self._last_sync_time:
            self.account.sync_last_sale_price(
                dt,
                self.data_portal,
            )
            self._last_sync_time = dt

    def on_dt_changed(self, dt: pd.Timestamp):
        self.datetime_manager.set_dt(dt)
        self.account.set_dt(dt)

    def set_slippage(self, equity_model=None, future_model=None):
        if self.initialized:
            raise ValueError("模型已经初始化")

        if equity_model is not None:
            self.account.slippage_models[ASSET_TYPE.Equity] = equity_model

        if future_model is not None:
            self.account.slippage_models[ASSET_TYPE.Future] = future_model

    def set_commission(self, equity_model=None, future_model=None):
        if self.initialized:
            raise ValueError("模型已经初始化")

        if equity_model is not None:
            self.account.commission_models[ASSET_TYPE.Equity] = equity_model

        if future_model is not None:
            self.account.commission_models[ASSET_TYPE.Future] = future_model

    def set_data_frequency(self, value):
        pass

    @property
    def data_frequency(self):
        pass

    def get_open_orders(self, asset=None):
        """
        Retrieve all of the current open get_orders
        """
        if asset is None:
            return {
                key: [order for order in orders]
                for key, orders in self.account.orders_tracker.open_orders.items()
                if orders
            }
        if asset in self.account.orders_tracker.open_orders:
            orders = self.account.orders_tracker.open_orders[asset]
            return [order for order in orders]
        return []

    def get_order(self, order_id):
        pass

    def cancel_order(self, order_param):
        """
        Parameters
        ----------
        order_param: order_id or Order object
        """
        order_id = order_param
        if isinstance(order_param, Order):
            order_id = order_param.id
        self.account.cancel(order_id)

    # the interface of data_portal
    def history(self, asset, field, bar_count, frequency):
        return self.data_portal.history_bars(asset, field, bar_count, frequency)

    def __repr__(self):
        template = """
{class_name}(
capital_base={capital_base},
sim_params={sim_params},
initialized={initialized},
slippage_models={slippage_models},
commission_models={commission_models},
account={account}
)
        """.strip()
        return template.format(
            class_name=self.__class__.__name__,
            capital_base=self.sim_params.capital_base,
            sim_params=repr(self.sim_params),
            initialized=self.initialized,
            slippage_models=repr(self.account.slippage_models),
            commission_models=repr(self.account.commission_models),
            account=repr(self.account),
        )


class AlgorithmIterator(DateTimeMixin):
    """
    AlgoIterator并不是一个迭代器, 而是Iterable, 你可以把他看作sim_engine的增强
    """
    EMISSION_TO_PERF_KEY_MAP = {"minute": "minute_perf", "daily": "daily_perf"}

    def __init__(
            self,
            algo: TradingAlgorithm,
            datetime_manager: DateTimeManager,
            sim_params: SimulationParameters,
            data_portal: DataPortal,
            sim_engine: typing.Union[DailySimulationClock, MinuteSimulationClock],
            benchmark_source=0,
            restrictions=None,
    ):
        self.algo = algo
        DateTimeMixin.__init__(self, datetime_manager)
        self.data_portal = data_portal
        self.current_data = data_portal
        self.restrictions = restrictions
        self.sim_params = sim_params
        self.sim_engine = sim_engine
        self.benchmark_source = benchmark_source

    def on_bar(self, dt: pd.Timestamp):
        # 15:00:00, order today, transact tomorrow
        self.set_dt(dt)
        self.algo.account.update_ledger()

        blotter = self.algo.account.orders_tracker
        new_transactions, new_commissions, closed_orders = blotter.get_transactions(self.current_data)
        blotter.prune_orders(closed_orders)

        for transaction in new_transactions:
            self.algo.account.handle_transaction(transaction)
            order = blotter.data[transaction.order_id]
            self.algo.account.handle_order(order)  # process order仅仅只是把订单放到对应的数据结构中

        for commission in new_commissions:
            self.algo.account.handle_commission(commission)

        self.algo._handle_data()  # 这里可能会下单了

        new_orders = blotter.get_new_orders()
        blotter.sweep_new_orders()

        # 将new_order放入对应的数据结构中
        for new_order in new_orders:
            self.algo.account.handle_order(new_order)

    def on_session_start(self, dt: pd.Timestamp):
        """
        """
        # 00:00:00
        self.set_dt(dt)

        self.algo.metrics_tracker.handle_start_of_session(
            dt,
            self.algo.data_portal,
        )

    def on_session_end(self, dt):
        """
        执行订单取消政策
        """
        # execute_order_cancellation_policy()
        # algo.validate_account_controls()
        # TODO: 这里更新了账户信息, 想想怎么分离
        perf_message = self.algo.metrics_tracker.handle_end_of_session(
            dt,
            self.data_portal,
        )
        perf_message["daily_perf"]["recorded_vars"] = self.algo.recorded_vars
        return perf_message

    def on_before_trading_start(self, dt):
        self.set_dt(dt)
        # self.algo.on_dt_changed(dt)
        self.algo._before_trading_start()

    def on_exit(self):
        """
        Remove references to algo, data portal, et al to break cycles and ensure
        deterministic cleanup of these objects when the simulation finishes.
        """
        self.algo.analyze()
        self.algo = None
        self.benchmark_source = None
        self.current_data = None
        self.data_portal = None

    def __iter__(self):

        if self.algo.data_frequency == "minute":
            raise NotImplementedError
            # def execute_order_cancellation_policy():
            #     self.algo.blotter.execute_cancel_policy(SESSION_END)
            #
            # def calculate_minute_capital_changes(dt):
            #     # process any capital changes that came between the last
            #     # and current minutes
            #     return self.algo.calculate_capital_changes(
            #         dt, emission_rate=self.emission_rate, is_interday=False
            #     )

        elif self.algo.data_frequency == "daily":

            def execute_order_cancellation_policy():
                # self.algo.account.execute_daily_cancel_policy(EVENT_TYPE.SESSION_END)
                # TODO: implement it
                return None
            def calculate_minute_capital_changes(dt):
                return []

        else:

            def execute_order_cancellation_policy():
                pass

            def calculate_minute_capital_changes(dt):
                return []

        event_bus = iter(self.sim_engine)
        while True:
            try:
                dt, event = next(event_bus)
                if event == EVENT_TYPE.BAR:
                    self.on_bar(dt)

                elif event == EVENT_TYPE.SESSION_START:
                    self.on_session_start(dt)

                elif event == EVENT_TYPE.SESSION_END:
                    yield self.on_session_end(dt)

                elif event == EVENT_TYPE.MINUTE_START:
                    raise NotImplementedError

                elif event == EVENT_TYPE.MINUTE_END:
                    raise NotImplementedError

                elif event == EVENT_TYPE.BEFORE_TRADING_START:
                    self.on_before_trading_start(dt)

            except StopIteration:
                break

        print("\n")
        self.on_exit()

    def _cleanup_expired_assets(self, dt, position_assets):
        """
        Clear out any assets that have expired before starting a new sim day.

        Performs two functions:

        1. Find all assets for which we have open get_orders and clears any get_orders whose aasets are
        on or after their auto_close_date.

        2. Find all assets for which we have positions and generates close_position events for any assets that have
        reached their auto_close_date.
        """
        pass

    def history(
            self,
            asset: Asset,
            fields: typing.Union[str, typing.Iterable[str]],
            bar_count: int,
            frequency: DATA_FREQUENCIES,
    ):
        return self.data_portal.history_bars(asset, fields, bar_count, frequency)





import numpy as np
import pandas as pd
import torchqtm.op.functional as F
from typing import Optional
import matplotlib.pyplot as plt
from scipy import stats


class RiskEvaluator(object):
    def __init__(self,
                 returns: pd.Series,
                 weights: pd.DataFrame = None,
                 freq=None,
                 rfr=0,
                 benchmark=None):
        self.returns = returns
        self.weights = weights
        self.freq = freq
        self.rfr = rfr
        self.benchmark = benchmark
        self.perf = {}

    def _sharpe(self):
        return sharpe(self.returns, self.freq, self.rfr)

    def _max_draw_down(self):
        series, val, longest_time = drawdown(self.returns)
        return series, val, longest_time

    def _turn_over(self):
        assert self.weights is not None
        return turnover(self.weights)

    def _beta(self):
        assert self.benchmark is not None
        pass

    def _alpha(self):
        assert self.benchmark is not None
        pass

    def _monthly_returns(self):
        pass

    def _yearly_returns(self):
        pass

    def _annualized_mean(self):
        pass

    def _annualized_volatility(self):
        pass


class FactorEvaluator(object):
    def __init__(self,
                 factor_scores: pd.DataFrame,
                 forward_returns: pd.DataFrame):
        self.factor_scores = factor_scores
        self.forward_returns = forward_returns

    def _ic(self):
        return ic(self.factor_scores, self.forward_returns)


def ic(factor_scores: pd.DataFrame, forward_returns: pd.DataFrame, method='pearson') -> pd.Series:
    """
    :return: id series
    """
    assert factor_scores.shape == forward_returns.shape
    rlt = F.cs_corr(factor_scores, forward_returns, method=method)
    rlt.fillna(0, inplace=True)
    return rlt


def drawdown(returns: pd.Series) -> (pd.Series, float, int):
    """
    :returns : drawdown series, max drawdown, longest drawdown
    """
    net_curve = (1+returns).cumpord()
    cmax = net_curve.cumax()
    rlt = cmax / net_curve - 1
    longest_drawdown = cmax.value_counts().max() - 1
    if isinstance(returns, pd.Series):
        rlt = pd.Series(rlt, index=returns.index)
    elif isinstance(returns, np.ndarray):
        rlt = np.array(rlt)
    return rlt, rlt.max(), longest_drawdown


def turnover(weights: pd.DataFrame) -> pd.Series:
    """
    calculate the amount traded at each rebalance rawdata as fraction of portfolio size.
    """
    weights.fillna(0, inplace=True)
    forward_weights = weights.shift(1)
    forward_weights.fillna(0, inplace=True)
    rlt = (weights - forward_weights).abs().sum(axis=1)
    return rlt


def sharpe(returns, freq=None, rfr=0):
    """
    Calculates the Sharpe ratio from a returns time series and the risk-free rate.

    Parameters
    ----------
    period_returns : DataSeries
        pandas series indexed by date with returns for each period as fraction of previous value
        (i.e. not in percent)
    freq : {'A', 'Q', 'M', 'D', or None}, optional, default None
        The frequency of the rawdata. If not None will annualize the result by multiplying by
        :math:`\sqrt{N}`
        where :math:`N` is the number of periods per year for the respective frequency
        (e.g. 252 for daily, 12 for monthly, 60 for weekly, etc.)
    rfr : float or pandas time series, default 0.
        Risk free rate, scaled to match returns rawdata frequency (i.e. annual if annual returns
        are provided, etc.)

    """
    if returns.isna().sum() > 0:
        raise ValueError('period_returns contains nan values.')

    mean = returns.mean()
    std = returns.std()
    _annf = freq
    return mean / std * _annf ** 0.5


def information_table(ic_data):
    ic_summary_table = pd.DataFrame()
    ic_summary_table["IC Mean"] = ic_data.mean()
    ic_summary_table["IC Std."] = ic_data.std()
    ic_summary_table["Risk-Adjusted IC"] = \
        ic_data.mean() / ic_data.std()
    t_stat, p_value = stats.ttest_1samp(ic_data, 0)
    ic_summary_table["t-stat(IC)"] = t_stat
    ic_summary_table["p-value(IC)"] = p_value
    ic_summary_table["IC Skew"] = stats.skew(ic_data)
    ic_summary_table["IC Kurtosis"] = stats.kurtosis(ic_data)










import torchqtm.op as op
import torchqtm.op.functional as F
import pandas as pd
import numpy as np

a = np.zeros((10, 20), dtype=np.float64)
a = pd.DataFrame(a)


def f(x):
    return np.mean(x)


a.rolling(5).apply(f)


# def algo():
#     lag = 22
#     ldf = {}
#     adjfactor = get_daily_data("AdjFactor",lines=lag)
#     for k in key:
#         ldf[k] = np.log(get_daily_data(k,lines=lag)*adjfactor)
#     ldf['ClosePrice_lag_1'] = np.apply_along_axis(lambda x:shift(x,1,cval=np.nan),0,ldf['ClosePrice'])
#     a = np.sqrt(np.nanmean(0.5*(ldf['HighPrice']-ldf['LowPrice'])**2,axis=0)- \
#         (2*np.log(2)-1)*np.nanmean((ldf['ClosePrice']-ldf['OpenPrice'])**2,axis=0)+ \
#         np.nanmean((ldf['OpenPrice']-ldf['ClosePrice_lag_1'])**2,axis=0))
#
#     return a


class Ross(op.Volatility):
    def __init__(self):
        super().__init__()

    def forward(self):
        raw_shape = self.env.Open.shape
        Open = np.array(F.log(self.env.Open), dtype=np.float64).reshape(*raw_shape, 1)
        High = np.array(F.log(self.env.High), dtype=np.float64).reshape(*raw_shape, 1)
        Low = np.array(F.log(self.env.Low), dtype=np.float64).reshape(*raw_shape, 1)
        Close = np.array(F.log(self.env.Close), dtype=np.float64).reshape(*raw_shape, 1)
        Closel1 = np.array(F.ts_delay(self.env.Close), dtype=np.float64).reshape(*raw_shape, 1)
        data = np.concatenate([Open, High, Low, Close, Closel1], axis=2)

        def aux_func(data_slice):
            cl = {
                'Open': 0,
                'High': 1,
                'Low': 2,
                'Close': 3,
                'Closel1': 4
            }
            return np.sqrt(np.nanmean(0.5 * (data_slice[..., cl['High']] - data_slice[..., cl['Low']]) ** 2, axis=0) -
                           (2 * np.log(2) - 1) * np.nanmean(
                (data_slice[..., cl['Close']] - data_slice[..., cl['Open']]) ** 2, axis=0) +
                           np.nanmean((data_slice[..., cl['Open']] - data_slice[..., cl['Closel1']]) ** 2, axis=0))
        rlt = F.ts_apply(data, 10, aux_func)
        return rlt



import torchqtm.op as op
import torchqtm.op.functional as F
from torchqtm.utils.rebalance import Monthly, Weekly, Daily
from torchqtm.utils.universe import StaticUniverse, IndexComponents
from torchqtm.utils.benchmark import BenchMark
from torchqtm.tdbt.backtest import GroupTester01, LongShort01


from torchqtm.base import BackTestEnv



import torchqtm.derivatives.option.functional as F
from torchqtm.utils import Timer

S = 100
K = 100
T = 1
r = 0.05
sigma = 0.2
price = 10.450583572185565  # Assuming the market price of option

# Pricing the option
price = F.bsm_call(T, S, K, r, sigma)
print("The Black-Scholes price is ", price)

# Calculating implied volatility
imp_vol = F.implied_vol(T, S, K, r, price, 'newton')
with Timer():
    print("The implied volatility is ", imp_vol)

# with Timer():
#     df = lambda x: (x-3)**2
#     ddf = lambda x: 2*(x-3)+1
#     tt = F._Newton(df, ddf, x_0=1, N=10000)[-1]
#     print(df(tt))




from distutils.core import setup
from setuptools import find_packages
from Cython.Build import cythonize
import numpy
import glob
import platform

PLATFORM = platform.system()


extensions = glob.glob("torchqtm/_C/*.pyx")
extensions.extend(glob.glob("torchqtm/window/*.pyx"))
extensions.extend(glob.glob("torchqtm/finance/*.pyx"))

install_requires = [
    'numpy',
    'pandas',
    'pandas_market_calendars',
    'cython',
]


def main():
    setup(
        ext_modules=cythonize(extensions, annotate=True),
        include_dirs=[numpy.get_include()],
        name="torchqtm",
        version="0.6.0",
        author="ny",
        author_email="nymath@163.com",
        install_requires=install_requires,
        description="None",
        long_description=open('README.md', 'r').read(),
        long_description_content_type="text/markdown",
        url="https://github.com/nymath/torchqtm",
        download_url="https://github.com/nymath/torchqtm/releases/tag/",
        packages=find_packages(),
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: Unix",
        ],
        python_requires='>=3.9',
    )


if __name__ == "__main__":
    main()

## setup cython
# python setup.py build_ext --inplace
## setup package
# pip install twine
# python setup.py sdist bdist_wheel
# twine upload dist/*
# twine upload --skip-existing dist/*

# ny.math
# xxx



from matplotlib import gridspec
import numpy as np
import pandas as pd
from torchqtm.edbt.temp.datahandler import BarData
from torchqtm.edbt.temp.backtest import BacktestEngine
from sklearn.linear_model import LinearRegression
import statsmodels.api as sm

lm = LinearRegression()
factor_name = 'fom123_mean'


class FixedPeriodRebalanceBacktester(BacktestEngine):

    def initialize(self):
        return None

    def handle_data(self):
        if self.current_bar < 22:
            print("有内鬼, 停止交易", self.current_bar)
            return None

        if self.bars.now in self.rebalance_date:
            print(self.now)
            account1 = self.get_account('account1')
            account2 = self.get_account('account2')
            num_of_stocks = 300
            status = self.history('status')
            cond_lst = status[status == 0].index  # 排除停牌
            factor_data = self.history('pb')
            float_mv = self.history('float_mv')
            ind_data = self.history('sector1')
            mymap = factor_data.groupby(ind_data).apply(np.mean)
            # 行业中性化
            factor_data = factor_data - ind_data.map(mymap).astype('float64')
            # 市值中性化
            data = pd.concat([factor_data, float_mv], axis=1)
            data.columns = ['factor_data', 'float_mv']
            data = data.dropna()
            data['intercept'] = 1
            model = sm.OLS(data['factor_data'], data[['intercept', 'float_mv']]).fit()
            residuals = data['factor_data'] - model.predict(data[['intercept', 'float_mv']])
            factor_data.loc[data.index] = residuals

            buy_lst = factor_data.nlargest(num_of_stocks).dropna().index
            buy_lst = buy_lst.intersection(cond_lst).to_list()
            buy_lst_2 = factor_data.nsmallest(num_of_stocks).dropna().index
            buy_lst_2 = buy_lst_2.intersection(cond_lst).to_list()
            if not buy_lst:
                return None
            weights = [1 / len(buy_lst)] * len(buy_lst)
            weights2 = [1 / len(buy_lst_2)] * len(buy_lst_2)
            account1.order_pct_to(buy_lst, weights)
            account2.order_pct_to(buy_lst_2, weights2)

        return None


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import pickle
    import copy

    # data = pd.DataFrame(np.random.normal(3000, 300, size=(1000, len(stocks))), columns=stocks,
    #                     index=pd.date_range('2000-01-01', periods=1000))
    # with open('./CICC/project1/datas.pkl', 'rb') as f:
    with open('quotes_dict.pkl', 'rb') as f:
        dfs = pickle.load(f)
    tempdata = copy.deepcopy(dfs['Close'])
    bar = BarData(dfs, start_date='20100101', end_date='20230401', lookback=60)
    # bar = ModifiedBarData(dfs, start_date='20100101', end_date='20230401', lookback=60)
    temp = FixedPeriodRebalanceBacktester(bar, ['account1', 'account2'])
    temp.run_strategy()
    rlt = temp.run_stats()

    fig = plt.figure(figsize=(20, 14), facecolor='white')
    gs = gridspec.GridSpec(4, 2, left=0.04, bottom=0.15, right=0.96, top=0.96, wspace=None, hspace=0,
                           height_ratios=[5, 1, 1, 1])
    ax = fig.add_subplot(gs[0, :])
    ax.plot(range(len(rlt['account1'])), (rlt['account1'] + 1).cumprod().values, color='red')
    ax.plot(range(len(rlt['account1'])), (rlt['account2'] + 1).cumprod().values, color='blue')

    ax.set_xlim(0, len(rlt['account1']))
    step = int(len(rlt['account1']) / 5)
    ax.set_xticks(range(0, len(rlt['account1']), step))
    ax.set_xticklabels(temp.bars.time_index.strftime('%Y-%m-%d')[::step])
    plt.title(f'{factor_name}')
    fig.show()
    # fig.savefig(f'./result/{factor_name}.pdf')



# distutils: language=c++
# https://cython.readthedocs.io/en/stable/src/userguide/numpy_tutorial.html
cimport numpy
import numpy as np
cimport cython
import pandas as pd
from torchqtm.config import __OP_MODE__
from sklearn.linear_model import LinearRegression

ctypedef fused ArrayType:
    float
    double
    long long



@cython.wraparound(False)
@cython.boundscheck(False)
def _regression_neut(Y, X):
    result = []
    for i in range(len(Y)):
        y = Y.values[i]
        y_demean = y - np.nanmean(y)
        x = X.values[i]
        x_demean = x - np.nanmean(x)
        residuals = y_demean - (np.nanmean(y_demean * x_demean) / np.nanvar(x_demean)) * x_demean
        result.append(residuals)
    return pd.DataFrame(np.array(result), index=Y.index, columns=Y.columns)


@cython.wraparound(False)
@cython.boundscheck(False)
def _regression_neuts(Y, others):
    result = []
    for i in range(len(Y)):
        y = Y.values[i]
        X = np.concatenate([x.values[i].reshape(-1, 1) for x in others], axis=1)
        model = LinearRegression()
        model.fit(X, y)
        y_pred = model.predict(X)
        residuals = y - y_pred
        result.append(residuals)
    return pd.DataFrame(np.array(result), index=Y.index, columns=Y.columns)


@cython.wraparound(False)
@cython.boundscheck(False)
def regression_neut(Y, others):
    assert isinstance(Y, pd.DataFrame)
    if not isinstance(others, list):
        return _regression_neut(Y, others)
    else:
        if __OP_MODE__ == "STABLE":
            return _regression_neuts(Y, others)
        else:
            pass

# roll_apply.pyx
cimport cython
import numpy as np









# blotter is the abstraction for the collection of get_orders
# 用于控制订单的整体行为

from abc import ABCMeta, abstractmethod
from torchqtm.finance.orders.order import Order
from collections import defaultdict
import pandas as pd
import logging
from typing import Dict, List, Optional, DefaultDict, Union, Any, Type, Tuple
from copy import copy

from torchqtm.finance.models.commission import CommissionModel, EquityCommissionModel, FutureCommissionModel
from torchqtm.finance.models.slippage import SlippageModel, EquitySlippageModel, FutureSlippageModel
from torchqtm.finance.models.cancel_policy import CancelPolicy, NeverCancel
from torchqtm.assets import Asset, Equity, Future
from torchqtm.data.data_portal import DataPortal
from torchqtm.finance.transaction import Transaction
from torchqtm.finance.models.commission import (
    # DEFAULT_PER_CONTRACT_COST,
    # FUTURE_EXCHANGE_FEES_BY_SYMBOL,
    PerContract,
    PerShare,
)
from torchqtm.finance.models.slippage import (
    DEFAULT_FUTURE_VOLUME_SLIPPAGE_BAR_LIMIT,
    DEFAULT_PER_CONTRACT_COST,
    FUTURE_EXCHANGE_FEES_BY_SYMBOL,
    VolatilityVolumeShare,
    FixedBasisPointsSlippage,
)

from torchqtm.types import ASSET_TYPE, EVENT_TYPE
from torchqtm.utils.datetime_utils import DateTimeManager, DateTimeMixin

log = logging.getLogger("Blotter")
warning_logger = logging.getLogger("AlgoWarning")


class OrdersTracker(DateTimeMixin):
    def __init__(
            self,
            datetime_manager: DateTimeManager = None,
            equity_slippage: SlippageModel = None,
            future_slippage: SlippageModel = None,
            equity_commission: CommissionModel = None,
            future_commission: CommissionModel = None,
            cancel_policy: CancelPolicy = None,
    ):

        DateTimeMixin.__init__(self, datetime_manager)
        self.cancel_policy = cancel_policy

        # these get_orders are aggregated by asset
        self.open_orders: DefaultDict[Asset, List[Order]] = defaultdict(list)

        # keep a dict of get_orders by their own id
        self.data: Dict[str, Order] = {}

        # holding get_orders that have come in since the last event.
        self.new_orders: List[Order] = []
        self.max_shares = int(1e11)
        # Here we use type to indicate that the key is Asset not the instance of Asset
        # TODO
        self.slippage_models: Dict[ASSET_TYPE, SlippageModel] = {
            ASSET_TYPE.Equity: equity_slippage or FixedBasisPointsSlippage(),
            ASSET_TYPE.Future: future_slippage
                               or VolatilityVolumeShare(
                volume_limit=DEFAULT_FUTURE_VOLUME_SLIPPAGE_BAR_LIMIT,
            ),
        }
        self.commission_models: Dict[ASSET_TYPE, CommissionModel] = {
            ASSET_TYPE.Equity: equity_commission or PerShare(),
            ASSET_TYPE.Future: future_commission
                               or PerContract(
                cost=DEFAULT_PER_CONTRACT_COST,
                exchange_fee=FUTURE_EXCHANGE_FEES_BY_SYMBOL,
            ),
        }

    def __repr__(self):
        raw_expression = f"""
{self.__class__.__name__}(
    slippage_models={self.slippage_models},
    commission_models={self.commission_models},
    open_orders={self.open_orders},
    get_orders={self.data},
    new_orders={self.new_orders},
    current_dt={self.current_dt}
)
        """.strip()
        return raw_expression

    def order(
            self,
            asset: Asset,
            amount: int,
            stop: float = None,
            limit: float = None,
            order_id: str = None,
    ) -> Optional[str]:

        if amount == 0:
            return None
        elif amount > self.max_shares:
            raise OverflowError("Can't order more than %d shares" % self.max_shares)

        order = Order(
            dt=self.current_dt,
            asset=asset,
            amount=amount,
            stop=stop,
            limit=limit,
            id=order_id,
        )

        self.open_orders[order.asset].append(order)
        self.data[order.id] = order
        self.new_orders.append(order)
        return order.id

    def cancel(
            self,
            order_id: str,
            relay_status: bool = True,
    ) -> None:
        if order_id not in self.data:
            return None

        current_order = self.data[order_id]

        if current_order.open:
            order_list = self.open_orders[current_order.asset]
            if current_order in order_list:
                order_list.remove(current_order)

            if current_order in self.new_orders:
                self.new_orders.remove(current_order)
            current_order.cancel()  # change the status to canceled
            current_order.dt = self.current_dt

            if relay_status:
                self.new_orders.append(current_order)

    def cancel_all_orders_for_asset(
            self,
            asset: Asset,
            warn: bool = False,
            relay_status: bool = True,
    ) -> None:
        """
        Cancel all open get_orders for a given asset
        """
        orders = self.open_orders[asset]
        for order in orders:
            if warn:
                # Message appropriately depending on whether there's
                # been a partial fill or not.
                if order.filled > 0:
                    warning_logger.warning(
                        "Your order for {order_amt} shares of "
                        "{order_sym} has been partially filled. "
                        "{order_filled} shares were successfully "
                        "purchased. {order_failed} shares were not "
                        "filled by the end of day and "
                        "were canceled.".format(
                            order_amt=order.amount,
                            order_sym=order.asset.symbol,
                            order_filled=order.filled,
                            order_failed=order.amount - order.filled,
                        )
                    )
                elif order.filled < 0:
                    warning_logger.warning(
                        "Your order for {order_amt} shares of "
                        "{order_sym} has been partially filled. "
                        "{order_filled} shares were successfully "
                        "sold. {order_failed} shares were not "
                        "filled by the end of day and "
                        "were canceled.".format(
                            order_amt=order.amount,
                            order_sym=order.asset.symbol,
                            order_filled=-1 * order.filled,
                            order_failed=-1 * (order.amount - order.filled),
                        )
                    )
                else:
                    warning_logger.warning(
                        "Your order for {order_amt} shares of "
                        "{order_sym} failed to fill by the end of day "
                        "and was canceled.".format(
                            order_amt=order.amount,
                            order_sym=order.asset.symbol,
                        )
                    )

        assert not orders
        del self.open_orders[asset]

    # End of day cancel for daily frequency
    def execute_daily_cancel_policy(self, event: EVENT_TYPE) -> None:
        if self.cancel_policy.should_cancel(event):
            warn = self.cancel_policy.warn_on_cancel
            for asset in copy(self.open_orders):
                orders = self.open_orders[asset]
                if len(orders) > 1:
                    # 为什么只取消了第0个订单
                    order = orders[0]
                    self.cancel(order.id, relay_status=True)
                    if warn:
                        if order.filled > 0:
                            warning_logger.warning(
                                "Your order for {order_amt} shares of "
                                "{order_sym} has been partially filled. "
                                "{order_filled} shares were successfully "
                                "purchased. {order_failed} shares were not "
                                "filled by the end of day and "
                                "were canceled.".format(
                                    order_amt=order.amount,
                                    order_sym=order.asset.symbol,
                                    order_filled=order.filled,
                                    order_failed=order.amount - order.filled,
                                )
                            )
                        elif order.filled < 0:
                            warning_logger.warning(
                                "Your order for {order_amt} shares of "
                                "{order_sym} has been partially filled. "
                                "{order_filled} shares were successfully "
                                "sold. {order_failed} shares were not "
                                "filled by the end of day and "
                                "were canceled.".format(
                                    order_amt=order.amount,
                                    order_sym=order.asset.symbol,
                                    order_filled=-1 * order.filled,
                                    order_failed=-1 * (order.amount - order.filled),
                                )
                            )
                        else:
                            warning_logger.warning(
                                "Your order for {order_amt} shares of "
                                "{order_sym} failed to fill by the end of day "
                                "and was canceled.".format(
                                    order_amt=order.amount,
                                    order_sym=order.asset.symbol,
                                )
                            )

    def execute_cancel_policy(self, event: EVENT_TYPE) -> None:
        if self.cancel_policy.should_cancel(event):
            warn = self.cancel_policy.warn_on_cancel
            for asset in copy(self.open_orders):
                self.cancel_all_orders_for_asset(asset, warn, relay_status=False)

    def reject(self, order_id: str, reason: str = "") -> None:
        """
        Mark the given order as 'rejected', which is functionally similar to
        cancelled. The distinction is that rejections are involuntary (and
        usually include a message from a broker indicating why the order was
        rejected) while cancels are typically user-driven.
        """
        if order_id not in self.data:
            return None

        current_order = self.data[order_id]
        order_list = self.open_orders[current_order.asset]
        if current_order in order_list:
            order_list.remove(current_order)

        if current_order in self.new_orders:
            self.new_orders.remove(current_order)

        current_order.reject(reason=reason)
        current_order.dt = self.current_dt

        self.new_orders.append(current_order)

    def hold(self, order_id: str, reason: str = "") -> None:
        """
        Mark the order with order_id as 'held'. Held is functionally similar
        to 'open'. When a fill (full or partial) arrives, the status
        will automatically change back to open/filled as necessary.
        """
        if order_id not in self.data:
            return None

        cur_order = self.data[order_id]
        if cur_order.open:
            if cur_order in self.new_orders:
                self.new_orders.remove(cur_order)
            cur_order.hold(reason=reason)
            cur_order.dt = self.current_dt
            # we want this order's new status to be relayed out
            # along with newly placed get_orders.
            self.new_orders.append(cur_order)

    def handle_splits(self, splits: List[Tuple[Asset, float]]) -> None:
        """
        Parameters
        ----------
        splits: list
            a list of splits. Each split is a tuple of (asset, ratio)
        """
        for asset, ratio in splits:
            if asset not in self.open_orders:
                continue

            orders_to_modify = self.open_orders[asset]
            for order in orders_to_modify:
                order.handle_split(ratio)

    def get_transactions(self, data: DataPortal) -> Tuple[List, List, List]:
        """Create a list of get_transactions based on the current open get_orders,
        slippage models, and commission models.

        Returns
        -------
        closed_orders: list
        commissions_list: list
        transactions_list: list
        """
        closed_orders: List[Order] = []
        commissions: List[Dict[str, Union[Asset, Order, float]]] = []
        transactions: List[Transaction] = []

        if self.open_orders:
            for asset, asset_orders in self.open_orders.items():
                slippage_model = self.slippage_models[asset.type]

                # 注意一下, 这里的order是asset_orders中的一个reference, 所以order状态的改变将直接体现在asset_orders中
                try:
                    it = iter(slippage_model.simulate(data, asset, asset_orders))
                except TypeError:
                    return transactions, commissions, closed_orders
                while True:
                    try:
                        order, txn = next(it)
                        commission_model = self.commission_models[asset.type]
                        additional_commission = commission_model.calculate(order, txn)

                        if additional_commission > 0:
                            commissions.append(
                                {
                                    "asset": order.asset,
                                    "order": order,
                                    "cost": additional_commission
                                }
                            )

                        order.filled += txn.amount
                        order.commission += additional_commission

                        order.dt = txn.dt

                        transactions.append(txn)

                        if not order.open:
                            closed_orders.append(order)
                    except StopIteration:
                        break
        return transactions, commissions, closed_orders

    def get_new_orders(self):
        return self.new_orders

    def sweep_new_orders(self):
        self.new_orders = []

    def prune_orders(self, closed_orders: List[Order]) -> None:
        """
        Remove all given get_orders from the blotter's open_orders list
        """
        for order in closed_orders:
            asset = order.asset
            asset_orders = self.open_orders[asset]
            try:
                asset_orders.remove(order)
            except ValueError:
                continue

        for asset in list(self.open_orders.keys()):
            if len(self.open_orders[asset]) == 0:
                del self.open_orders[asset]



import logging
from collections import defaultdict

from torchqtm.data.data_portal import DataPortal
from torchqtm.finance.account import Account
from torchqtm.finance.metrics.metric import Metric
from torchqtm.utils.exchange_calendar import XSHGExchangeCalendar
import numpy as np
import pandas as pd
import sys
from torchqtm.types import DATA_FREQUENCIES

log = logging.getLogger(__name__)
import typing


def print_progress(progress):
    '''
    progress 是一个介于0和1之间的浮点数，代表当前进度
    '''
    bar_length = 50  # 进度条的长度，你可以根据自己的需求进行修改
    assert 0 <= progress <= 1

    # 计算已完成部分和未完成部分的长度
    completed_length = int(bar_length * progress)
    uncompleted_length = bar_length - completed_length

    # 创建进度条字符串
    progress_bar = '[' + '#' * completed_length + '-' * uncompleted_length + ']'

    # 在同一行打印进度条和进度百分比
    sys.stdout.write("\r" + progress_bar + ' {:.2f}%'.format(progress * 100))
    sys.stdout.flush()


class MetricsTracker:
    """The algorithm's interface to the registered risk and performance metrics

    """

    def __init__(
            self,
            trading_calendar: XSHGExchangeCalendar,
            first_session: pd.Timestamp,
            last_session: pd.Timestamp,
            capital_base: float,
            emission_rate: DATA_FREQUENCIES,
            data_frequency: DATA_FREQUENCIES,
            account: Account,
            metrics: typing.Iterable[Metric],
    ):
        assert isinstance(account, Account), "avoid creating a new account"

        self._benchmark_source = None
        self.emission_rate: typing.Literal["daily", "minute"] = emission_rate
        self._trading_calendar = trading_calendar
        self._first_session: pd.Timestamp = first_session
        self._last_session: pd.Timestamp = last_session
        self._capital_base: float = capital_base
        self._metrics = metrics

        self._current_session = first_session
        self._market_open, self._market_close = self._execution_open_and_close(
            trading_calendar,
            first_session,
        )

        self._session_count = 0
        self._sessions = trading_calendar.sessions_in_range(
            first_session,
            last_session,
        )

        self._total_session_count = len(self._sessions)
        self._account = account

    def _execution_open_and_close(self, trading_calendar, session):
        # TODO: implement this function
        return pd.Timestamp("20050104 09:31:00"), pd.Timestamp("20050104 15:00:00")

    def _progress(self):
        if self.emission_rate == "minute":
            return 1.0

        else:
            return self._session_count / self._total_session_count

    def start_of_simulation(
            self,
            account: Account,
            emission_rate: typing.Literal["daily", "minute"],
            trading_calendar,
            sessions: pd.DatetimeIndex,
            benchmark_source,
    ):
        for metric in self._metrics:
            metric.start_of_simulation(
                account=account,
                emission_rate=emission_rate,
                trading_calendar=trading_calendar,
                sessions=sessions,
                benchmark_source=benchmark_source,
            )

    def end_of_simulation(
            self,
            packet: typing.Dict,
            account: Account,
            trading_calendar,
            sessions: pd.DatetimeIndex,
            data_portal: DataPortal,
            benchmark_source,
    ):
        for metric in self._metrics:
            metric.end_of_simulation(
                packet=packet,
                account=account,
                trading_calendar=trading_calendar,
                sessions=sessions,
                data_portal=data_portal,
                benchmark_source=benchmark_source,
            )

    def start_of_session(
            self,
            account: Account,
            session: pd.Timestamp,
            data_portal: DataPortal,
    ):
        for metric in self._metrics:
            metric.start_of_session(
                account=account,
                session=session,
                data_portal=data_portal,
            )

    def end_of_session(
            self,
            packet: typing.Dict,
            account: Account,
            session: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        for metric in self._metrics:
            metric.end_of_session(
                packet=packet,
                account=account,
                session=session,
                session_idx=session_idx,
                data_portal=data_portal,
            )

    def start_of_bar(self, *args, **kwargs):
        for metric in self._metrics:
            metric.start_of_bar(*args, **kwargs)

    def end_of_bar(
            self,
            packet: typing.Dict,
            account: Account,
            dt: pd.Timestamp,
            session_idx: int,
            data_portal: DataPortal,
    ):
        for metric in self._metrics:
            metric.end_of_bar(
                packet=packet,
                account=account,
                dt=dt,
                session_idx=session_idx,
                data_portal=data_portal,
            )

    def handle_start_of_simulation(self, benchmark_source):
        self._benchmark_source = benchmark_source
        self.start_of_simulation(
            self._account,
            self.emission_rate,
            self._trading_calendar,
            self._sessions,
            benchmark_source,
        )

    def handle_start_of_session(self, session_label: pd.Timestamp, data_portal: DataPortal):
        self._account.start_of_session(session_label)

        # TODO: load and process dividends
        # adjustment_reader = data_portal.adjustment_reader
        # if adjustment_reader is not None:
        #     # this is None when running with a dataframe source
        #     ledger.process_dividends(
        #         session_label,
        #         self._asset_finder,
        #         adjustment_reader,
        #     )
        self._current_session = session_label
        self._market_open, self._market_close = self._execution_open_and_close(
            self._trading_calendar,
            session_label,
        )

        self.start_of_session(
            account=self._account,
            session=session_label,
            data_portal=data_portal,
        )

        # handle any splits that impact any positions or any open orders.
        assets_we_care_about = (
                self._account.positions.keys() | self._account.orders_tracker.open_orders.keys()
        )

        if assets_we_care_about:
            splits = data_portal.get_splits(assets_we_care_about, session_label)
            if splits:
                # 订单要进行分割, metrics_tracker也能进行分割???? 其中谁发挥了作用
                self._account.orders_tracker.handle_splits(splits)
                self._account.handle_splits(splits)

    def handle_end_of_session(self, dt, data_portal):
        """Handles the close of the given day.

        Parameters
        ----------
        dt : Timestamp
            The most recently completed simulation datetime.
        data_portal : DataPortal
            The current data portal.

        Returns
        -------
        A daily perf packet.
        """
        completed_session = self._current_session

        if self.emission_rate == "daily":
            # this method is called for both minutely and daily emissions, but
            # this chunk of code here only applies for daily emissions. (since
            # it's done every minute, elsewhere, for minutely emission).
            self._account.sync_last_sale_price(dt, data_portal)

        session_ix = self._session_count
        # increment the day counter before we move markers forward.
        self._session_count += 1

        packet = {
            "period_start": self._first_session,
            "period_end": self._last_session,
            "capital_base": self._capital_base,
            "daily_perf": {
                "period_open": self._market_open,
                "period_close": dt,
            },
            "cumulative_perf": {
                "period_open": self._first_session,
                "period_close": self._last_session,
            },
            "progress": self._progress(),
            "cumulative_risk_metrics": {},
        }
        self._account.end_of_session(session_ix)
        self.end_of_session(
            packet,
            self._account,
            completed_session,
            session_ix,
            data_portal,
        )
        print_progress(self._progress())
        return packet

    def handle_simulation_end(self, data_portal: DataPortal):
        """When the simulation is complete, run the full period risk report
        and send it out on the results socket.
        """
        log.info(
            "Simulated %(days)s trading days\n first open: %(first)s\n last close: %(last)s",
            dict(
                days=self._session_count,
                first=self._trading_calendar.session_open(self._first_session),
                last=self._trading_calendar.session_close(self._last_session),
            ),
        )

        packet = {}
        self.end_of_simulation(
            packet,
            self._account,
            self._trading_calendar,
            self._sessions,
            data_portal,
            self._benchmark_source,
        )
        print_progress(1)
        return packet



# distutils: language=c++


cimport cython
# from cython cimport PyObject

cimport numpy as cnp
import numpy as np
import pandas as pd
from torchqtm.finance.positions.position import Position
from torchqtm.assets import Future

cdef class PositionStats:
    """Compute values from the current positions

    Attributes
    ----------
    """
    cdef cnp.float64_t gross_exposure
    cdef cnp.float64_t gross_value
    cdef cnp.float64_t long_exposure
    cdef cnp.float64_t long_value
    cdef cnp.float64_t net_exposure
    cdef cnp.float64_t net_value
    cdef cnp.float64_t short_exposure
    cdef cnp.float64_t short_value
    cdef cnp.uint64_t longs_count
    cdef cnp.uint64_t shorts_count
    cdef object position_exposure_array
    cdef object position_exposure_series

    cdef object underlying_value_array
    cdef object underlying_index_array

    @classmethod
    def new(cls):
        cdef PositionStats self = cls()
        self.position_exposure_series = es = pd.Series(
            np.array([], dtype='float64'),
            index=np.array([], dtype='int64'),
        )
        self.underlying_value_array = self.position_exposure_array = es.values
        self.underlying_index_array = es.index.values
        return self


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef calculate_positions_tracker_stats(positions, PositionStats stats):
    """Calculate various stats about the current positions
    
    Parameters
    ----------
    positions : OrderedDict
    stats : PositionStats
    
    Returns
    -------
    position_stats : PositionStats
        The computed statistics
    """
    cdef Py_ssize_t n_pos = len(positions)
    cdef cnp.ndarray[cnp.uint64_t] index
    cdef cnp.ndarray[cnp.float64_t] position_exposure

    cdef cnp.ndarray[cnp.int64_t] old_index = stats.underlying_index_array
    cdef cnp.ndarray[cnp.float64_t] old_position_exposure = stats.underlying_index_array

    cdef cnp.float64_t value
    cdef cnp.float64_t exposure

    cdef cnp.float64_t net_value
    cdef cnp.float64_t gross_value
    cdef cnp.float64_t long_value = 0.0
    cdef cnp.float64_t short_value = 0.0

    cdef cnp.float64_t net_exposure
    cdef cnp.float64_t gross_exposure
    cdef cnp.float64_t long_exposure = 0.0
    cdef cnp.float64_t short_exposure = 0.0

    cdef cnp.uint64_t longs_count = 0
    cdef cnp.uint64_t shorts_count = 0


    # if len(old_index) < n_pos:
    #     # we don't have enough space in the cached buffer, allocate a new
    #     # array
    #     stats.underlying_index_array = index = np.empty(n_pos, dtype='int64')
    #     stats.underlying_value_array = position_exposure = np.empty(
    #         n_pos,
    #         dtype='float64',
    #     )
    #
    #     stats.position_exposure_array = position_exposure
    #     # create a new series to expose the arrays
    #     stats.position_exposure_series = pd.Series(
    #         position_exposure,
    #         index=index,
    #     )
    # elif len(old_index) > n_pos:
    #     # we have more space than needed, slice off the extra but leave it
    #     # available
    #     index = old_index[:n_pos]
    #     position_exposure = old_position_exposure[:n_pos]
    #
    #     stats.position_exposure_array = position_exposure
    #     # create a new series with the sliced arrays
    #     stats.position_exposure_series = pd.Series(
    #         position_exposure,
    #         index=index,
    #     )
    # else:
    #     # we have exactly the right amount of space, no slicing or allocation
    #     # needed
    #     index = old_index
    #     position_exposure = old_position_exposure
    #
    #     stats.position_exposure_array = position_exposure
    #     stats.position_exposure_series = pd.Series(
    #         position_exposure,
    #         index=index,
    #     )

    # cdef InnerPosition position
    cdef Py_ssize_t ix = 0

    for position in positions.values():
        # NOTE: this loop does a lot of stuff!
        # we call this function every time the portfolio value is needed,
        # which is at least once per simulation day, so let's not iterate
        # through every single position multiple times.
        exposure = position.amount * position.last_sale_price

        if type(position.asset) is Future:
            # Futures don't have an inherent position value.
            value = 0

            # unchecked cast, this is safe because we do a type check above
            exposure *= position.asset.price_multiplier
        else:
            value = exposure

        if exposure > 0:
            longs_count += 1
            long_value += value
            long_exposure += exposure
        elif exposure < 0:
            shorts_count += 1
            short_value += value
            short_exposure += exposure

        # index[ix] = position.asset.sid
        # position_exposure[ix] = exposure
        # ix += 1

    net_value = long_value + short_value
    gross_value = long_value - short_value

    net_exposure = long_exposure + short_exposure
    gross_exposure = long_exposure - short_exposure

    stats.gross_exposure = gross_exposure
    stats.gross_value = gross_value
    stats.long_exposure = long_exposure
    stats.long_value = long_value
    stats.longs_count = longs_count
    stats.net_exposure = net_exposure
    stats.net_value = net_value
    stats.short_exposure = short_exposure
    stats.short_value = short_value
    stats.shorts_count = shorts_count



import pandas as pd
import typing

# import talib

from torchqtm.edbt.algorithm import TradingAlgorithm
from torchqtm.edbt.sim_params import SimulationParameters

from torchqtm.finance.account import Account
from torchqtm.finance.metrics.tracker import MetricsTracker
from torchqtm.utils.calendar_utils import get_calendar
from torchqtm.utils.datetime_utils import DateTimeManager
from torchqtm.types import DATA_FREQUENCIES
from torchqtm.data.data_portal import DataPortal
from torchqtm.assets import Equity
from torchqtm.finance.metrics.loader import DEFAULT_METRICS


# Design a time machine


class run_algo(object):
    def __init__(
            self,
            Algo: typing.Type[TradingAlgorithm],
            data_frequency: DATA_FREQUENCIES,
            capital_base: float,
            bundle: str,
            start: pd.Timestamp,
            end: pd.Timestamp,
            output,
            trading_calendar,
            metrics_set,
            local_namespace,
            environ,
            account_configs,
            benchmark_spec,
    ):

        if trading_calendar is None:
            trading_calendar = get_calendar("XSHG")
        datetime_manager = DateTimeManager(trading_calendar)

        # benchmark_symbol, benchmark_returns = benchmark_spec.resolve(start, end)
        benchmark_symbol, benchmark_returns = (0, 0)

        data_portal = DataPortal(
            data_frequency=data_frequency,
            trading_calendar=datetime_manager.trading_calendar,
            datetime_manager=datetime_manager,
            restrictions=None,
        )

        sim_params = SimulationParameters(
            start_session=start,
            end_session=end,
            trading_calendar=trading_calendar,
            capital_base=capital_base,
            data_frequency=data_frequency,
        )

        account = Account(
            datetime_manager=datetime_manager,
            trading_sessions=trading_calendar.sessions_in_range(start, end),
            capital_base=capital_base,
            data_frequency=data_frequency,
        )

        metrics_tracker = MetricsTracker(
            trading_calendar=trading_calendar,
            first_session=start,
            last_session=end,
            capital_base=capital_base,
            emission_rate=data_frequency,
            data_frequency=data_frequency,
            account=account,
            metrics=metrics_set,
        )

        algo = Algo(
            sim_params=sim_params,
            data_portal=data_portal,
            namespace=None,
            trading_calendar=datetime_manager.trading_calendar,
            datetime_manager=datetime_manager,
            benchmark_returns=benchmark_returns,
            account=account,
            metrics_tracker=metrics_tracker,
        )

        it = iter(algo.get_generator())
        while True:
            try:
                next(it)
            except StopIteration:
                break


class TestAlgo(TradingAlgorithm):
    def initialize(self):
        self.safe_set_attr("sym", Equity("000001.XSHE"))
        self.safe_set_attr("count", 0)
        self.sym: Equity
        self.count: int

    def before_trading_start(self):
        pass

    def handle_data(self):
        # for i in range(300):
        #     prices = self.history(self.sym, 'close', 21, "daily")
        #
        # short_avg = talib.SMA(prices, 5)
        # long_avg = talib.SMA(prices, 20)
        #
        # current_position = self.account.get_position(self.sym)
        # if short_avg[-1] - long_avg[-1] < 0 and short_avg[-2] - long_avg[-2] > 0 and current_position.amount > 0:
        #     self.order(self.sym, -100)
        #
        # if short_avg[-1] - long_avg[-1] > 0 and short_avg[-2] - long_avg[-2] < 0:
        #     for i in range(100):
        #         self.order(self.sym, 10)
        pass

    def analyze(self):
        print(self.account.ledger.cash)
        print(self.account.ledger.portfolio_value)


import sys


def print_progress(progress):
    '''
    progress 是一个介于0和1之间的浮点数，代表当前进度
    '''
    bar_length = 50  # 进度条的长度，你可以根据自己的需求进行修改
    assert 0 <= progress <= 1

    # 计算已完成部分和未完成部分的长度
    completed_length = int(bar_length * progress)
    uncompleted_length = bar_length - completed_length

    # 创建进度条字符串
    progress_bar = '[' + '#' * completed_length + '-' * uncompleted_length + ']'

    # 在同一行打印进度条和进度百分比
    sys.stdout.write("\r" + progress_bar + ' {:.2f}%'.format(progress * 100))
    sys.stdout.flush()


if __name__ == "__main__":
    Algo = TestAlgo
    data_frequency: DATA_FREQUENCIES = "daily"
    capital_base = 1e6
    bundle = "rqalpha"
    start = pd.Timestamp("20160101")
    end = pd.Timestamp("20220101")
    output = None
    trading_calendar = None
    metrics_set = DEFAULT_METRICS
    local_namespace = None
    environ = None
    account_configs = None
    benchmark_spec = None

    run_algo(
        Algo=Algo,
        data_frequency=data_frequency,
        capital_base=capital_base,
        bundle=bundle,
        start=start,
        end=end,
        output=output,
        trading_calendar=trading_calendar,
        metrics_set=metrics_set,
        local_namespace=local_namespace,
        environ=environ,
        account_configs=account_configs,
        benchmark_spec=benchmark_spec,
    )



import uuid
import math
from enum import Enum
from sys import float_info
import pandas as pd
from typing import Dict, Tuple, Optional
from torchqtm.assets import Asset
from torchqtm.types import ORDER_STATUS, ORDER_TYPE
import torchqtm.finance.zp_math as zp_math

SELL = 1 << 0
BUY = 1 << 1
STOP = 1 << 2
LIMIT = 1 << 3

ORDER_FIELDS_TO_IGNORE = {"type", "direction", "_status", "asset"}


def consistent_round(val):
    if (val % 1) >= 0.5:
        return math.ceil(val)
    else:
        return round(val)


def asymmetric_round_price(price, prefer_round_down, tick_size, diff=0.95):
    precision = zp_math.number_of_decimal_places(tick_size)
    multiplier = int(tick_size * (10**precision))
    diff -= 0.5  # shift the difference down
    diff *= 10**-precision  # adjust diff to precision of tick size
    diff *= multiplier  # adjust diff to value of tick_size

    # Subtracting an epsilon from diff to enforce the open-ness of the upper
    # bound on buys and the lower bound on sells.  Using the actual system
    # epsilon doesn't quite get there, so use a slightly less epsilon-ey value.
    epsilon = float_info.epsilon * 10
    diff = diff - epsilon

    # relies on rounding half away from zero, unlike numpy's bankers' rounding
    rounded = tick_size * consistent_round(
        (price - (diff if prefer_round_down else -diff)) / tick_size
    )
    if zp_math.tolerant_equals(rounded, 0.0):
        return 0.0
    return rounded


# TODO: create a base class for Order
class Order:
    # using __slot__ to save on memory usage.
    __slots__ = [
        "id",
        "dt",
        "reason",
        "created",
        "asset",
        "amount",
        "filled",
        "commission",
        "_status",
        "stop",
        "limit",
        "stop_reached",
        "limit_reached",
        "direction",
        "type",
        "broker_order_id",
    ]

    def __init__(
            self,
            dt: pd.Timestamp,
            asset: Asset,
            amount: int,
            stop: Optional[float] = None,
            limit: Optional[float] = None,
            filled: int = 0,
            commission: float = .0,
            id: Optional[str] = None,
    ):
        self.id = self.make_id() if id is None else id
        self.dt = dt
        self.reason = None
        self.created = dt
        self.asset = asset
        self.amount = amount
        self.filled = filled
        self.commission = commission
        self._status = ORDER_STATUS.OPEN
        self.type = self.get_order_type(stop, limit)
        # then, we need to modify the stop, limit price
        self.stop, self.limit = self.get_stop_limit_price(stop, limit, self.amount > 0)
        self.stop_reached = False
        self.limit_reached = False
        self.direction = math.copysign(1, self.amount)
        self.broker_order_id = None

    @staticmethod
    def make_id() -> str:
        return uuid.uuid4().hex

    @staticmethod
    def get_order_type(stop_price: Optional[float], limit_price: Optional[float]):
        if limit_price and stop_price:
            return ORDER_TYPE.STOP_LIMIT
        elif not limit_price and stop_price:
            return ORDER_TYPE.STOP
        elif limit_price and not stop_price:
            return ORDER_TYPE.LIMIT
        else:
            return ORDER_TYPE.MARKET

    @staticmethod
    def get_stop_limit_price(
            stop: Optional[float],
            limit: Optional[float],
            is_buy: Optional[float],
    ) -> Tuple[Optional[float], Optional[float]]:
        if not limit and not stop:
            # for market order
            true_stop = None
            true_limit = None
        elif limit and not stop:
            # for limit order
            true_stop = None
            true_limit = asymmetric_round_price(limit, is_buy, tick_size=0.01)
        elif not limit and stop:
            # for stop order
            # why here we should use not is_buy
            true_stop = asymmetric_round_price(stop, not is_buy, tick_size=0.01)
            true_limit = None
        else:
            # for stop limit order
            true_stop = asymmetric_round_price(stop, not is_buy, tick_size=0.01)
            true_limit = asymmetric_round_price(limit, is_buy, tick_size=0.01)

        return true_stop, true_limit

    def to_dict(self) -> Dict[str, object]:
        dct = {
            name: getattr(self, name)
            for name in self.__slots__
            if name not in ORDER_FIELDS_TO_IGNORE
        }

        if self.broker_order_id is None:
            del dct["broker_order_id"]

        dct['status'] = self.status
        return dct

    def check_trigger(self, price: float, dt: pd.Timestamp) -> None:
        """
        handle_transaction internal state based on price triggers and the trade event's price.
        """
        stop_reached, limit_reached, sl_stop_reached = self.check_order_triggers(price)
        if (stop_reached, limit_reached) != (self.stop_reached, self.limit_reached):
            self.dt = dt
        self.stop_reached = stop_reached
        self.limit_reached = limit_reached
        if sl_stop_reached:
            # Change the STOP LIMIT order into a LIMIT order
            self.stop = None

    def check_order_triggers(self, current_price: float) -> Tuple[bool, bool, bool]:
        """
        return (stop_reached, limit_reached)
        - `market order`: (False, False)
        - `stop order`: (~, False)
        - `limit order`: (False, ~)
        - `stop limit order`: (~, ~)
        """
        if self.triggered:
            return self.stop_reached, self.limit_reached, False

        stop_reached = False
        limit_reached = False
        sl_stop_reached = False

        order_type = 0

        if self.amount > 0:
            order_type |= BUY
        else:
            order_type |= SELL

        if self.stop is not None:
            order_type |= STOP

        if self.limit is not None:
            order_type |= LIMIT

        if order_type == BUY | STOP | LIMIT:
            if current_price >= self.stop:
                sl_stop_reached = True
                if current_price <= self.limit:
                    limit_reached = True
        elif order_type == SELL | STOP | LIMIT:
            if current_price <= self.stop:
                sl_stop_reached = True
                if current_price >= self.limit:
                    limit_reached = True
        elif order_type == BUY | STOP:
            if current_price >= self.stop:
                stop_reached = True
        elif order_type == SELL | STOP:
            if current_price <= self.stop:
                stop_reached = True
        elif order_type == BUY | LIMIT:
            if current_price <= self.limit:
                limit_reached = True
        elif order_type == SELL | LIMIT:
            # This is a SELL LIMIT order
            if current_price >= self.limit:
                limit_reached = True

        return stop_reached, limit_reached, sl_stop_reached

    def handle_split(self, ratio: float) -> None:
        """
        handle_transaction the amount, limit_price and stop_price
        思考这样的做法是不是后复权呢
        """
        self.amount = int(self.amount / ratio)

        if self.limit is not None:
            self.limit = round(self.limit * ratio, 2)

        if self.stop is not None:
            self.stop = round(self.stop * ratio, 2)

    def set_status(self, status: ORDER_STATUS) -> None:
        self._status = status

    def cancel(self) -> None:
        self._status = ORDER_STATUS.CANCELLED

    def reject(self, reason="") -> None:
        self._status = ORDER_STATUS.REJECTED
        self.reason = reason

    def hold(self, reason="") -> None:
        self._status = ORDER_STATUS.HELD
        self.reason = reason

    @property
    def status(self):
        if not self.open_amount:
            return ORDER_STATUS.FILLED
        elif self._status == ORDER_STATUS.HELD and self.filled:
            return ORDER_STATUS.OPEN
        else:
            return self._status

    @property
    def open(self):
        return self.status in [ORDER_STATUS.OPEN, ORDER_STATUS.HELD]

    @property
    def triggered(self) -> bool:
        """
        for a market order, True
        for a stop order, True iff stop_reached
        for a limit order, True iff limit_reached
        """
        if self.stop is not None and not self.stop_reached:
            return False

        if self.limit is not None and not self.limit_reached:
            return False

        return True

    @property
    def open_amount(self) -> int:
        return self.amount - self.filled

    def __repr__(self):
#         template = """
# {class_name}(
#     id={id},
#     dt={dt},
#     asset={asset},
#     amount={amount},
#     stop={stop},
#     limit={limit},
#     filled={filled},
#     status={status},
# )
#         """
        return f"Order({self.to_dict().__repr__()})"

